# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining writer factory."""

from __future__ import annotations

import csv
import os
import warnings
from importlib import import_module

from gaarf.io.writers import abs_writer


def create_writer(
  writer_option: str, **kwargs: str
) -> type[abs_writer.AbsWriter]:
  """Factory function for creating concrete writer.

  Writer is created based on `writer_option` and possible `kwargs` needed
  to correctly instantiate it.

  Args:
      writer_option: Type of writer.
      kwargs: Any possible arguments needed o instantiate writer.

  Returns:
      Concrete instantiated writer.
  """
  if writer_option in ('bq', 'bigquery'):
    writer_module = import_module('gaarf.io.writers.bigquery_writer')
    return writer_module.BigQueryWriter(**kwargs)
  if writer_option == 'sqldb':
    writer_module = import_module('gaarf.io.writers.sqldb_writer')
    return writer_module.SqlAlchemyWriter(**kwargs)
  if writer_option in ('sheet', 'sheets'):
    writer_module = import_module('gaarf.io.writers.sheets_writer')
    return writer_module.SheetWriter(**kwargs)
  if writer_option == 'console':
    writer_module = import_module('gaarf.io.writers.console_writer')
    return writer_module.ConsoleWriter(**kwargs)
  if writer_option == 'csv':
    writer_module = import_module('gaarf.io.writers.csv_writer')
    return writer_module.CsvWriter(**kwargs)
  if writer_option == 'json':
    writer_module = import_module('gaarf.io.writers.json_writer')
    return writer_module.JsonWriter(**kwargs)
  return import_module('gaarf.io.writers.null_writer').NullWriter(writer_option)


class WriterFactory:
  """Deprecated class for creating concrete writer."""

  def __init__(self) -> None:
    warnings.warn(
      'WritingFactory is deprecated, use `gaarf.io.writer.create_writer` '
      'function instead',
      category=DeprecationWarning,
      stacklevel=2,
    )

  def create_writer(
    self, writer_option: str, **kwargs: str
  ) -> type[abs_writer.AbsWriter]:
    """Factory method for creating concrete writer.

    Writer is created based on `writer_option` and possible `kwargs` needed
    to correctly instantiate it.

    Args:
        writer_option: Type of writer.
        kwargs: Any possible arguments needed o instantiate writer.

    Returns:
        Concrete instantiated writer.
    """
    return create_writer(writer_option, **kwargs)


# Deprecated writers
class BigQueryWriter:
  """Deprecated class for creating BigQueryWriter."""

  def __new__(cls, project: str, dataset: str, location: str = 'US', **kwargs):
    warnings.warn(
      'Loading BigQueryWriter from `gaarf.io.writer` is deprecated; '
      'Import BigQueryWriter from `gaarf.io.writers.bigquery_writer` '
      'instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return create_writer(
      'bq', project=project, dataset=dataset, location=location
    )


class CsvWriter:
  """Deprecated class for creating CsvWriter."""

  def __new__(
    cls,
    destination_folder: str = os.getcwd(),
    delimiter: str = ',',
    quotechar: str = '"',
    quoting=csv.QUOTE_MINIMAL,
    **kwargs,
  ):
    warnings.warn(
      'Loading CsvWriter from `gaarf.io.writer` is deprecated; '
      'Import CsvWriter from `gaarf.io.writers.csv_writer` instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return create_writer(
      'csv',
      destination_folder=destination_folder,
      delimiter=delimiter,
      quotechar=quotechar,
      quoting=quoting,
    )


class SqlAlchemyWriter:
  """Deprecated class for creating SqlAlchemyWriter."""

  def __new__(cls, connection_string, if_exists='replace', **kwargs):
    warnings.warn(
      'Loading SqlAlchemyWriter from `gaarf.io.writer` is deprecated; '
      'Import SqlAlchemyWriter from `gaarf.io.writers.sqldb_writer` '
      'instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return create_writer(
      'sqldb', connection_string=connection_string, if_exists=if_exists
    )


class SheetWriter:
  """Deprecated class for creating SheetWriter."""

  def __new__(
    cls,
    share_with,
    credentials_file,
    spreadsheet_url=None,
    is_append=False,
    **kwargs,
  ):
    warnings.warn(
      'Loading SheetWriter from `gaarf.io.writer` is deprecated; '
      'Import SheetWriter from `gaarf.io.writers.sheets_writer` instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return create_writer(
      'sheet',
      share_with=share_with,
      credentials_file=credentials_file,
      spreadsheet_url=spreadsheet_url,
      is_append=is_append,
    )


class StdoutWriter:
  """Deprecated class for creating StdoutWriter."""

  def __new__(cls, page_size: int = 10, **kwargs):
    warnings.warn(
      'Loading StdoutWriter from `gaarf.io.writer` is deprecated; '
      'Import ConsoleWriter from `gaarf.io.writers.console_writer` '
      'instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return create_writer('console', page_size=page_size)


class ZeroRowException(Exception):
  """Raised when report has no data."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import Dict

from smart_open import open  # type: ignore


class AbsReader(abc.ABC):
  @abc.abstractmethod
  def read(self, query_path: str, **kwargs):
    raise NotImplementedError


class FileReader(AbsReader):
  def __init__(self):
    pass

  def read(self, query_path, **kwargs):
    with open(query_path, 'r') as f:
      raw_query_text = f.read()
    return raw_query_text


class ConsoleReader(AbsReader):
  def __init__(self):
    pass

  def read(self, query_path, **kwargs):
    return query_path


class NullReader(AbsReader):
  def __init__(self, reader_option, **kwargs):
    raise ValueError(f'{reader_option} is unknown reader type!')

  def read(self):
    print('Unknown reader type!')


class ReaderFactory:
  reader_options: Dict[str, AbsReader] = {}

  def __init__(self):
    self.load_reader_options()

  def load_reader_options(self):
    self.reader_options['file'] = FileReader
    self.reader_options['console'] = ConsoleReader

  def create_reader(self, reader_option: str, **kwargs) -> AbsReader:
    if reader_option in self.reader_options:
      return self.reader_options[reader_option](**kwargs)
    else:
      return NullReader(reader_option)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import proto
"""Defines queries that are accessible via `builtin` namespace.

Built-in queries provides quick access to important elements in Google Ads API
with a simple SQL syntax.
All built-in queries are available in `builtin` namespace
(i.e. `builtin.ocid_mapping`) and can be added to SQL scripts
(i.e. `SELECT * FROM builtin.ocid_mapping`).

"""

from __future__ import annotations

import functools
import operator
import re
from collections import abc, defaultdict
from typing import Callable

import gaarf


def get_ocid_mapping(
  report_fetcher: gaarf.report_fetcher.AdsReportFetcher, accounts: list[str]
) -> gaarf.report.GaarfReport:
  """Returns mapping between external customer_id and OCID parameter.


  OCID parameter is used to build links to Google Ads entities in UI.

  Args:
      report_fetcher: An instance of AdsReportFetcher.
      accounts: Google Ads accounts to get data on OCID mapping.

  Returns:
      Report with mapping between external customer_id and OCID parameter.
  """
  query = (
    'SELECT customer.id AS account_id, '
    'metrics.optimization_score_url  AS url FROM customer'
  )
  mapping = []
  for account in accounts:
    if report := report_fetcher.fetch(query, account):
      for row in report:
        if ocid := re.findall(r'ocid=(\w+)', row.url):
          mapping.append([row.account_id, ocid[0]])
          break
      if not ocid:
        mapping.append([int(account), '0'])
    else:
      mapping.append([int(account), '0'])
  return gaarf.report.GaarfReport(
    results=mapping,
    column_names=[
      'account_id',
      'ocid',
    ],
  )


def get_account_hierarchy_flattened(
  report_fetcher: gaarf.report_fetcher.AdsReportFetcher, accounts: list[str]
) -> gaarf.report.GaarfReport:
  SEED_ACCOUNTS_QUERY = """
    SELECT
        customer_client.level AS level,
        customer_client.manager AS is_manager,
        customer_client.id AS account_id
    FROM customer_client
    WHERE customer_client.level <= 1
    """

  NESTED_ACCOUNTS_QUERY = """
    SELECT
        customer.id AS mcc_id,
        customer.descriptive_name AS mcc_name,
        {level} AS level,
        customer_client_link.client_customer~0 AS account_id
    FROM customer_client_link
    """

  seed_accounts = (
    accounts if isinstance(accounts, abc.MutableSequence) else [accounts]
  )
  hierarchy = report_fetcher.fetch(SEED_ACCOUNTS_QUERY, seed_accounts)
  level_mapping = defaultdict(list)
  for row in hierarchy:
    if row.is_manager:
      level_mapping[row.level].append(row.account_id)

  reports: list[gaarf.report.GaarfReport] = []
  for level, accounts in level_mapping.items():
    if accounts:
      report_fetcher.customer_ids = accounts
      reports.append(
        report_fetcher.fetch(NESTED_ACCOUNTS_QUERY.format(level=level))
      )

  combined_report = functools.reduce(operator.add, reports)
  df = combined_report.to_pandas()
  max_level = max(df.level)
  root_mcc_df = df[df.level == 0][['mcc_id', 'mcc_name', 'account_id']]
  root_mcc_df = root_mcc_df.rename(
    columns={'mcc_id': 'root_mcc_id', 'mcc_name': 'root_mcc_name'}
  )
  root_mcc_df['account_id'] = root_mcc_df['account_id'].astype(int)
  selectable_columns = ['root_mcc_id', 'root_mcc_name']
  if max_level == 0:
    return gaarf.report.GaarfReport.from_pandas(root_mcc_df)
  else:
    for i in range(1, max(df.level) + 1):
      if i == 1:
        n_mcc_df = df[df.level == i][['mcc_id', 'mcc_name', 'account_id']]
        mcc_ids = n_mcc_df.merge(
          root_mcc_df, how='right', left_on='mcc_id', right_on='account_id'
        )[
          [
            'root_mcc_id',
            'root_mcc_name',
            'account_id_x',
            'mcc_id',
            'mcc_name',
            'account_id_y',
          ]
        ]
      else:
        previous_level = i - 1
        n_minus_1_mcc_df = df[df.level == previous_level][
          ['mcc_id', 'mcc_name', 'account_id']
        ]
        n_minus_1_mcc_df = n_minus_1_mcc_df.rename(
          columns={
            'mcc_id': f'level_{previous_level}_mcc_id',
            'mcc_name': f'level_{previous_level}_mcc_name',
          }
        )
        mcc_ids = n_mcc_df.merge(
          n_minus_1_mcc_df, how='right', left_on='mcc_id', right_on='account_id'
        )[
          [
            f'level_{previous_level}_mcc_id',
            f'level_{previous_level}_mcc_name',
            'account_id_x',
            'mcc_id',
            'mcc_name',
            'account_id_y',
          ]
        ]

      mcc_ids.mcc_name.fillna('Direct', inplace=True)
      mcc_ids.account_id_x.fillna(mcc_ids.account_id_y, inplace=True)
      mcc_ids.mcc_id.fillna(mcc_ids.account_id_y, inplace=True)
      mcc_ids = mcc_ids.rename(
        columns={
          'mcc_name': f'level_{i}_mcc_name',
          'account_id_x': 'account_id',
          'mcc_id': f'level_{i}_mcc_id',
        }
      )[
        selectable_columns
        + [f'level_{i}_mcc_id', f'level_{i}_mcc_name', 'account_id']
      ]
      mcc_ids[f'level_{i}_mcc_id'] = mcc_ids[f'level_{i}_mcc_id'].astype(int)
    mcc_ids['account_id'] = mcc_ids['account_id'].astype(int)
    return gaarf.report.GaarfReport.from_pandas(mcc_ids)


BUILTIN_QUERIES: dict[str, Callable] = {
  'ocid_mapping': get_ocid_mapping,
  'account_hierarchy_flattened': get_account_hierarchy_flattened,
}



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for executing Gaarf queries and writing them to local/remote.


Module defines two major classes:
    * AdsReportFetcher - to perform fetching data from Ads API, parsing it
      and returning GaarfReport.
    * AdsQueryExecutor - to perform fetching data from Ads API in a form of
      GaarfReport and saving it to local/remote storage.
"""

from __future__ import annotations

import warnings
from collections.abc import Sequence

from gaarf import api_clients, executors, report_fetcher


class AdsReportFetcher:
  """Deprecated class for creating AdsReportFetcher."""

  def __new__(
    cls,
    api_client: api_clients.BaseClient,
    customer_ids: Sequence[str] | None = None,
  ) -> report_fetcher.AdsReportFetcher:
    warnings.warn(
      'Loading AdsReportFetcher from `gaarf.query_executor` is '
      'deprecated; Import AdsReportFetcher from `gaarf.report_fetcher` '
      'instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return report_fetcher.AdsReportFetcher(api_client, customer_ids)


class AdsQueryExecutor:
  """Deprecated class for creating AdsQueryExecutor."""

  def __new__(
    cls,
    api_client: api_clients.BaseClient,
  ) -> executors.ads_executor.AdsQueryExecutor:
    warnings.warn(
      'Loading AdsReportFetcher from `gaarf.query_executor` is '
      'deprecated; Import AdsReportFetcher from `gaarf.report_fetcher` '
      'instead',
      category=DeprecationWarning,
      stacklevel=2,
    )
    return executors.ads_executor.AdsQueryExecutor(api_client)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest

from gaarf.io import writer


@pytest.mark.parametrize('option', ['bq', 'bigquery'])
def test_create_writer_returns_correct_fields_for_bq_option(option):
  bq_writer = writer.create_writer(
    option, project='fake_project', dataset='fake_dataset'
  )
  assert bq_writer.dataset_id == 'fake_project.fake_dataset'


def test_create_writer_returns_correct_fields_for_csv_option():
  csv_writer = writer.create_writer('csv', destination_folder='/fake_folder')
  assert csv_writer.destination_folder == '/fake_folder'


def test_create_writer_returns_correct_fields_for_sheets_option():
  sheet_writer = writer.create_writer(
    'sheet',
    share_with='1@google.com',
    credentials_file='home/me/client_secret.json',
  )
  assert sheet_writer.share_with == '1@google.com'
  assert sheet_writer.credentials_file == 'home/me/client_secret.json'


def test_create_writer_returns_correct_fields_for_sqldb_option():
  sqlalchemy_writer = writer.create_writer(
    'sqldb', connection_string='protocol://user:password@host:port/db'
  )
  assert sqlalchemy_writer.connection_string == (
    'protocol://user:password@host:port/db'
  )


def test_create_writer_returns_correct_fields_for_json_option():
  json_writer = writer.create_writer('json', destination_folder='/fake_folder')
  assert json_writer.destination_folder == '/fake_folder'


def test_null_writer_raises_unknown_writer_error():
  with pytest.raises(ValueError):
    writer.create_writer('non-existing-option')


class TestDeprecatedWriters:
  def test_instantiating_writer_factory_is_deprected(self):
    with pytest.deprecated_call():
      writer.WriterFactory()

  def test_instantiating_console_writer_is_deprected(self):
    with pytest.warns(DeprecationWarning) as w:
      writer.StdoutWriter()
      assert len(w) == 1
      assert str(w[0].message) == (
        'Loading StdoutWriter from `gaarf.io.writer` is deprecated; '
        'Import ConsoleWriter from `gaarf.io.writers.console_writer` '
        'instead'
      )

  def test_instantiating_csv_writer_is_deprected(self):
    with pytest.warns(DeprecationWarning) as w:
      writer.CsvWriter()
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)
      assert str(w[0].message) == (
        'Loading CsvWriter from `gaarf.io.writer` is deprecated; '
        'Import CsvWriter from `gaarf.io.writers.csv_writer` instead'
      )

  def test_instantiating_sqlalchemy_writer_is_deprected(self):
    with pytest.warns(DeprecationWarning) as w:
      writer.SqlAlchemyWriter(connection_string='fake-connection')
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)
      assert str(w[0].message) == (
        'Loading SqlAlchemyWriter from `gaarf.io.writer` is '
        'deprecated; '
        'Import SqlAlchemyWriter from `gaarf.io.writers.sqldb_writer` '
        'instead'
      )

  def test_instantiating_sheet_writer_is_deprected(self):
    with pytest.warns(DeprecationWarning) as w:
      writer.SheetWriter(share_with='fake-user', credentials_file=None)
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)
      assert str(w[0].message) == (
        'Loading SheetWriter from `gaarf.io.writer` is deprecated; '
        'Import SheetWriter from `gaarf.io.writers.sheets_writer` '
        'instead'
      )

  def test_instantiating_bigquery_writer_is_deprected(self):
    with pytest.warns(DeprecationWarning) as w:
      writer.BigQueryWriter(project='fake-project', dataset='fake-dataset')
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)
      assert str(w[0].message) == (
        'Loading BigQueryWriter from `gaarf.io.writer` is deprecated; '
        'Import BigQueryWriter from `gaarf.io.writers.bigquery_writer` '
        'instead'
      )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defing `gaarf-bq` CLI utility.

`gaarf-bq` allows to execute BigQuery queries based on Gaarf config.
"""

from __future__ import annotations

import argparse
import functools
from concurrent import futures

from gaarf import bq_executor
from gaarf.cli import utils
from gaarf.io import reader  # type: ignore


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument('query', nargs='+')
  parser.add_argument('-c', '--config', dest='gaarf_config', default=None)
  parser.add_argument('--project', dest='project')
  parser.add_argument(
    '--dataset-location', dest='dataset_location', default=None
  )
  parser.add_argument('--save-config', dest='save_config', action='store_true')
  parser.add_argument(
    '--no-save-config', dest='save_config', action='store_false'
  )
  parser.add_argument(
    '--config-destination', dest='save_config_dest', default='config.yaml'
  )
  parser.add_argument('--log', '--loglevel', dest='loglevel', default='info')
  parser.add_argument('--logger', dest='logger', default='local')
  parser.add_argument('--dry-run', dest='dry_run', action='store_true')
  parser.add_argument(
    '--parallel-queries', dest='parallel_queries', action='store_true'
  )
  parser.add_argument(
    '--no-parallel-queries', dest='parallel_queries', action='store_false'
  )
  parser.add_argument(
    '--parallel-threshold', dest='parallel_threshold', default=None
  )
  parser.set_defaults(save_config=False)
  parser.set_defaults(dry_run=False)
  parser.set_defaults(parallel_queries=True)
  args = parser.parse_known_args()
  main_args = args[0]

  logger = utils.init_logging(
    loglevel=main_args.loglevel.upper(), logger_type=main_args.logger
  )
  config = utils.ConfigBuilder('gaarf-bq').build(vars(main_args), args[1])
  logger.debug('config: %s', config)
  if main_args.save_config and not main_args.gaarf_config:
    utils.ConfigSaver(main_args.save_config_dest).save(config)
  if main_args.dry_run:
    exit()

  config = utils.initialize_runtime_parameters(config)
  logger.debug('initialized config: %s', config)

  bigquery_executor = bq_executor.BigQueryExecutor(
    project_id=config.project, location=config.dataset_location
  )
  bigquery_executor.create_datasets(config.params.get('macro'))

  reader_client = reader.FileReader()

  if main_args.parallel_queries:
    logger.info('Running queries in parallel')
    with futures.ThreadPoolExecutor(
      max_workers=main_args.parallel_threshold
    ) as executor:
      future_to_query = {
        executor.submit(
          bigquery_executor.execute,
          query,
          reader_client.read(query),
          config.params,
        ): query
        for query in sorted(main_args.query)
      }
      for future in futures.as_completed(future_to_query):
        query = future_to_query[future]
        utils.postprocessor_runner(query, future.result, logger)
  else:
    logger.info('Running queries sequentially')
    for query in sorted(main_args.query):
      callback = functools.partial(
        executor.execute, query, reader_client.read(query), config.params
      )
      utils.postprocessor_runner(query, callback, logger)


if __name__ == '__main__':
  main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for installing gaarf as a package."""

from __future__ import annotations

import itertools
import pathlib

import setuptools

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

EXTRAS_REQUIRE = {
  'pandas': [
    'pandas>=1.3.4',
  ],
  'sqlalchemy': [
    'sqlalchemy',
  ],
  'simulator': [
    'Faker',
  ],
  'sheets': [
    'gspread',
  ],
  'bq': [
    'google-cloud-bigquery',
    'pandas>=1.3.4',
    'smart_open[gcs]',
  ],
}
EXTRAS_REQUIRE['full'] = list(set(itertools.chain(*EXTRAS_REQUIRE.values())))

setuptools.setup(
  name='google-ads-api-report-fetcher',
  version='1.15.0',
  python_requires='>3.8',
  description=(
    'Library for fetching reports from Google Ads API '
    'and saving them locally & remotely.'
  ),
  long_description=README,
  long_description_content_type='text/markdown',
  url='https://github.com/google/ads-api-reports-fetcher',
  author='Google Inc. (gTech gPS CSE team)',
  author_email='no-reply@google.com',
  license='Apache 2.0',
  classifiers=[
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: Apache Software License',
  ],
  packages=setuptools.find_packages(),
  install_requires=[
    'google-ads>=25.0.0',
    'smart_open',
    'jinja2==3.1.4',
    'python-dateutil',
    'typing-extensions',
    'rich',
    'tenacity',
  ],
  extras_require=EXTRAS_REQUIRE,
  setup_requires=[
    'pytest-runner',
  ],
  tests_requires=[
    'pytest',
    'pytest-mock',
  ],
  entry_points={
    'console_scripts': [
      'gaarf=gaarf.cli.gaarf:main',
      'gaarf-py=gaarf.cli.gaarf:main',
      'gaarf-bq=gaarf.cli.bq:main',
      'gaarf-py-bq=gaarf.cli.bq:main',
      'gaarf-simulator=gaarf.cli.simulator:main',
      'gaarf-sql=gaarf.cli.sql:main',
    ]
  },
)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest

from gaarf import api_clients, query_executor


@pytest.fixture
def test_client(mocker, config_path):
  mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
  return api_clients.GoogleAdsApiClient(path_to_config=config_path)


class TestAdsReportFetcher:
  def test_instantiating_ads_report_fetcher_Is_deprecated(self, test_client):
    with pytest.warns(DeprecationWarning) as w:
      query_executor.AdsReportFetcher(api_client=test_client)
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)


class TestAdsQueryExecutor:
  def test_instantiating_ads_query_executor_Is_deprecated(self, test_client):
    with pytest.warns(DeprecationWarning) as w:
      query_executor.AdsQueryExecutor(api_client=test_client)
      assert len(w) == 1
      assert issubclass(w[0].category, DeprecationWarning)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for formatting Gaarf reports before writing."""

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Callable, Union, get_args

import proto  # type: ignore
from typing_extensions import TypeAlias

from gaarf.report import GaarfReport

_NESTED_FIELD: TypeAlias = Union[
  list,
  proto.marshal.collections.repeated.RepeatedComposite,
  proto.marshal.collections.repeated.Repeated,
]


class FormattingStrategy:
  """Interface for all formatting strategies applied to GaarfReport."""

  def apply_transformations(self, report: GaarfReport) -> GaarfReport:
    """Applies class transformation to report."""
    raise NotImplementedError

  def _cast_to_enum(self, enum: type[Enum], value: str | Enum) -> Enum:
    """Ensures that strings are always converted to Enums."""
    return enum[value.upper()] if isinstance(value, str) else value


class ArrayHandling(Enum):
  """Specifies acceptable options for ArrayHandlingStrategy."""

  STRINGS = 1
  ARRAYS = 2


class ArrayHandlingStrategy(FormattingStrategy):
  """Handles arrays in the report.

  Arrays can be left as-is or converted to strings with required delimiter.

  Attributes:
      type_: Type of array handling (ARRAYS, STRINGS).
      delimiter: Symbol used as delimiter when type_ is STRINGS.
  """

  def __init__(
    self,
    type_: ArrayHandling | str = ArrayHandling.STRINGS,
    delimiter: str = '|',
  ) -> None:
    """Initializes strategy based on type_ and delimiter.

    Args:
        type_: Type of array handling (ARRAYS, STRINGS).
        delimiter: Symbol used as delimiter when type_ is STRINGS.
    """
    self.type_ = self._cast_to_enum(ArrayHandling, type_)
    self.delimiter = delimiter

  def apply_transformations(self, report: GaarfReport) -> GaarfReport:
    """Replaces arrays in the report."""
    if self.type_ == ArrayHandling.ARRAYS:
      return report

    formatted_rows = self._format_rows(report.results, self._delimiter_join)
    formatted_placeholders = self._format_rows(
      report.results_placeholder, lambda x: ''
    )
    return GaarfReport(
      results=formatted_rows,
      column_names=report.column_names,
      results_placeholder=formatted_placeholders,
    )

  def _format_rows(
    self, rows: list[list], nested_field_handler: Callable
  ) -> list[list]:
    """Formats rows of report based on join_strategy.

    Args:
        rows: Rows on GaarfReport.
        nested_field_handler: Handlers to nested structures.

    Returns:
        Formatted rows.
    """
    formatted_rows = []
    for row in rows:
      formatted_row = []
      for field in row:
        if isinstance(field, get_args(_NESTED_FIELD)):
          field = nested_field_handler(field)
        formatted_row.append(field)
      formatted_rows.append(formatted_row)
    return formatted_rows

  def _delimiter_join(self, field: _NESTED_FIELD) -> str:
    """Helper function to perform join by an instance delimiter.

    Args:
        field: A nested field.

    Returns:
        The same field but concatenated to string.
    """
    return self.delimiter.join([str(element) for element in field])


def format_report_for_writing(
  report: GaarfReport, formatting_strategies: list[FormattingStrategy]
) -> GaarfReport:
  """Applies formatting strategies to report.

  Args:
      report: Report that needs to be formatted.
      formatting_strategies: Strategies to be applied to report.

  Returns:
      New report with updated data.
  """
  for strategy in formatting_strategies:
    report = strategy.apply_transformations(report)
  report.disable_scalar_conversions()
  return report


def format_extension(
  path_object: str, current_extension: str = '.sql', new_extension: str = ''
) -> str:
  """Formats query path to required extension.

  Args:
      path_object: Path to query.
      current_extension: Extension of the query.
      new_extension: Required extension

  Returns:
     Path with an updated extension.
  """
  path_object_name = Path(path_object).name
  if len(path_object_name.split('.')) > 1:
    return path_object_name.replace(current_extension, new_extension)
  return f'{path_object}{new_extension}'



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module defining key query elements for building and parsing GAQL query."""

from __future__ import annotations

import ast
import contextlib
import dataclasses
import datetime
import operator
import re
from typing import Generator

from dateutil import relativedelta

from gaarf import api_clients, exceptions, query_post_processor

VALID_VIRTUAL_COLUMN_OPERATORS = (
  ast.BinOp,
  ast.UnaryOp,
  ast.operator,
  ast.Num,
  ast.Expression,
)


@dataclasses.dataclass(frozen=True)
class VirtualColumn:
  """Represents element in Gaarf query that either calculated or plugged-in.

  Virtual columns allow performing basic manipulation with metrics and
  dimensions (i.e. division or multiplication) as well as adding raw text
  values directly into report.

  Attributes:
    type: Type of virtual column, either build-in or expression.
    value: Value of the field after macro expansion.
    fields: Possible fields participating in calculations.
    substitute_expression: Formatted expression.
  """

  type: str
  value: str
  fields: list[str] | None = None
  substitute_expression: str | None = None


@dataclasses.dataclass
class ExtractedLineElements:
  """Helper class for parsing query lines.

  Attributes:
    fields: All fields extracted from the line.
    alias: Optional alias assign to a field.
    virtual_column: Optional virtual column extracted from query line.
    customizer: Optional values for customizers associated with a field.
  """

  field: str | None
  alias: str | None
  virtual_column: VirtualColumn | None
  customizer: dict[str, str | int]


@dataclasses.dataclass
class ProcessedField:
  """Helper class to store fields with its customizers.

  Attributes:
    field: Extractable field.
    customizer_type: Type of customizer to be applied to the field.
    customizer_value: Value to be used in customizer.
  """

  field: str
  customizer_type: str | None = None
  customizer_value: int | str | None = None


@dataclasses.dataclass
class QueryElements:
  """Contains raw query and parsed elements.

  Attributes:
      query_title: Title of the query that needs to be parsed.
      query_text: Text of the query that needs to be parsed.
      resource_name: Name of Google Ads API reporting resource.
      fields: Ads API fields that need to be fetched.
      column_names: Friendly names for fields which are used when saving data
      customizers: Attributes of fields that need to be be extracted.
      virtual_columns: Attributes of fields that need to be be calculated.
      is_constant_resource: Whether resource considered a constant one.
      is_builtin_query: Whether query is built-in.
  """

  query_title: str
  query_text: str
  resource_name: str
  fields: list[str] | None = None
  column_names: list[str] | None = None
  customizers: dict[str, dict[str, str]] | None = None
  virtual_columns: dict[str, VirtualColumn] | None = None
  is_constant_resource: bool = False
  is_builtin_query: bool = False


class CommonParametersMixin:
  """Helper mixin to inject set of common parameters to all queries."""

  _common_params = {
    'date_iso': lambda: datetime.date.today().strftime('%Y%m%d'),
    'yesterday_iso': lambda: (
      datetime.date.today() - relativedelta.relativedelta(days=1)
    ).strftime('%Y%m%d'),
    'current_date': lambda: datetime.date.today().strftime('%Y-%m-%d'),
    'current_datetime': lambda: datetime.datetime.today().strftime(
      '%Y-%m-%d %H:%M:%S'
    ),
  }

  @property
  def common_params(self):
    """Instantiates common parameters to the current values."""
    return {key: value() for key, value in self._common_params.items()}


class QuerySpecification(
  CommonParametersMixin, query_post_processor.PostProcessorMixin
):
  """Simplifies fetching data from API and it's further parsing..

  Attributes:
    text: Query text.
    title: Query title.
    args: Optional parameters to be dynamically injected into query text.
    api_version: Version of Google Ads API.
  """

  def __init__(
    self,
    text: str,
    title: str | None = None,
    args: dict | None = None,
    api_version: str = api_clients.GOOGLE_ADS_API_VERSION,
  ) -> None:
    """Instantiates QuerySpecification based on text, title and optional args.

    Args:
      text: Query text.
      title: Query title.
      args: Optional parameters to be dynamically injected into query text.
      api_version: Version of Google Ads API.
    """
    self.text = text
    self.title = title
    self.args = args or {}
    self._api_version = api_version

  @property
  def base_client(self):
    """Helper for validating identified query fields."""
    return api_clients.BaseClient(self._api_version)

  @property
  def macros(self) -> dict[str, str]:
    """Returns macros with injected common parameters."""
    common_params = dict(self.common_params)
    if macros := self.args.get('macro'):
      common_params.update(macros)
    return common_params

  @property
  def expanded_query(self) -> str:
    """Applies necessary transformations to query."""
    query_text = self.expand_jinja(self.text, self.args.get('template'))
    query_lines = self._remove_comments_from_query(query_text)
    query_text = ' '.join(query_lines)
    try:
      return query_text.format(**self.macros).strip()
    except KeyError as e:
      raise exceptions.GaarfMacroException(
        f'No value provided for macro {str(e)}.'
      ) from e

  def generate(self) -> QueryElements:
    """Generates necessary query elements based on query text and arguments.

    Returns:
        Various elements parsed from a query (text, fields,
        column_names, etc).

    Raises:
        GaarfResourceException: If query contains invalid resource_name.
        GaarfMacroException: If missing values for one of the query macros.
    """
    resource_name = self._extract_resource_from_query()
    if is_builtin_query := bool(resource_name.startswith('builtin')):
      return QueryElements(
        query_title=resource_name.replace('builtin.', ''),
        query_text=self.expanded_query,
        fields=None,
        column_names=None,
        customizers=None,
        virtual_columns=None,
        resource_name=resource_name,
        is_constant_resource=False,
        is_builtin_query=True,
      )
    if not is_builtin_query and resource_name not in dir(
      self.base_client.google_ads_row
    ):
      raise exceptions.GaarfResourceException(
        f'Invalid resource specified in the query: {resource_name}'
      )
    is_constant_resource = bool(resource_name.endswith('_constant'))
    fields = []
    column_names = []
    customizers = {}
    virtual_columns = {}

    for line in self._extract_query_lines():
      line_elements = self._extract_line_elements(line)
      column_name = line_elements.alias
      if field := line_elements.field:
        fields.append(field)
      column_names.append(column_name)

      if virtual_column := line_elements.virtual_column:
        virtual_columns[column_name] = virtual_column
      if customizer := line_elements.customizer:
        customizers[column_name] = customizer
    return QueryElements(
      query_title=self.title,
      query_text=self._create_gaql_query(fields, virtual_columns),
      fields=fields,
      column_names=column_names,
      customizers=customizers,
      virtual_columns=virtual_columns,
      resource_name=resource_name,
      is_constant_resource=is_constant_resource,
      is_builtin_query=is_builtin_query,
    )

  def _create_gaql_query(
    self,
    fields: list[str],
    virtual_columns: dict[str, VirtualColumn],
  ) -> str:
    """Generate valid GAQL query.

    Based on original Gaarf query text, a set of field and virtual columns
    constructs new GAQL query to be sent to Ads API.

    Args:
        fields:
            All fields that need to be fetched from API.
        virtual_columns:
            Virtual columns that might contain extra fields for fetching.

    Returns:
        Valid GAQL query.
    """
    virtual_fields = [
      field
      for name, column in virtual_columns.items()
      if column.type == 'expression'
      for field in column.fields
    ]
    if virtual_fields:
      fields = fields + virtual_fields
    query_text = (
      f'SELECT {", ".join(fields)} '
      f'FROM {self._extract_resource_from_query()} '
      f'{self._extract_filters()}'
    )
    query_text = self._remove_trailing_comma(query_text)
    query_text = self._unformat_type_field_name(query_text)
    return re.sub(r'\s+', ' ', query_text).strip()

  def _remove_comments_from_query(self, query_text: str) -> list[str]:
    """Removes comments and converts text to lines."""
    result: list[str] = []
    for line in query_text.split('\n'):
      if re.match('^(#|--|//)', line):
        continue
      cleaned_query_line = re.sub(
        ';$', '', re.sub('(--|//).*$', '', line).strip()
      )
      result.append(cleaned_query_line)
    return result

  def _extract_resource_from_query(self) -> str:
    """Finds resource_name in query_text.

    Returns:
      Found resource.

    Raises:
      GaarfResourceException: If resource_name isn't found.
    """
    if resource_name := re.findall(
      r'FROM\s+([\w.]+)', self.expanded_query, flags=re.IGNORECASE
    ):
      return str(resource_name[0]).strip()
    raise exceptions.GaarfResourceException(
      f'No resource found in query: {self.expanded_query}'
    )

  def _extract_query_lines(self) -> Generator[str, None, None]:
    """Helper for extracting fields with aliases from query text.

    Yields:
      Line in query between SELECT and FROM statements.
    """
    selected_rows = re.sub(
      r'\bSELECT\b|FROM .*', '', self.expanded_query, flags=re.IGNORECASE
    ).split(',')
    for row in selected_rows:
      if non_empty_row := row.strip():
        yield non_empty_row

  def _extract_filters(self) -> str:
    if where_statement := re.search(
      ' (WHERE|LIMIT|ORDER BY|PARAMETERS) .+',
      self.expanded_query,
      re.IGNORECASE,
    ):
      return where_statement.group(0)
    return ''

  def _extract_line_elements(self, query_line: str) -> ExtractedLineElements:
    """Parses query line into elements.

    Args:
      query_line: Field name with optional alias.

    Returns:
      Parsed elements (field, alias, virtual_column).
    """
    field, *alias = re.split(' [Aa][Ss] ', query_line)
    processed_field = self._process_field(field)
    field = processed_field.field
    if self._is_valid_google_ads_field(field):
      virtual_column = None
    else:
      virtual_column = self._convert_to_virtual_column(field)
    if alias and processed_field.customizer_type:
      customizer = {
        'type': processed_field.customizer_type,
        'value': processed_field.customizer_value,
      }
    else:
      customizer = {}
    if virtual_column and not alias:
      raise exceptions.GaarfVirtualColumnException(
        'Virtual attributes should be aliased'
      )
    return ExtractedLineElements(
      field=self._format_type_field_name(field)
      if not virtual_column and field
      else None,
      alias=self._normalize_column_name(alias[0] if alias else field),
      virtual_column=virtual_column,
      customizer=customizer,
    )

  def _process_field(self, raw_field: str) -> ProcessedField:
    """Process field to extract possible customizers.

    Args:
        raw_field: Unformatted field string value.

    Returns:
        ProcessedField that contains formatted field with customizers.
    """
    raw_field = raw_field.replace(r'\s+', '').strip()
    if self._is_quoted_string(raw_field):
      return ProcessedField(field=raw_field)
    if len(resources := self._extract_resource_element(raw_field)) > 1:
      field_name, resource_index = resources
      return ProcessedField(
        field=field_name,
        customizer_type='resource_index',
        customizer_value=int(resource_index),
      )

    if len(nested_fields := self._extract_nested_resource(raw_field)) > 1:
      field_name, nested_field = nested_fields
      return ProcessedField(
        field=field_name,
        customizer_type='nested_field',
        customizer_value=nested_field,
      )
    if len(pointers := self._extract_pointer(raw_field)) > 1:
      field_name, pointer = pointers
      return ProcessedField(
        field=field_name, customizer_type='pointer', customizer_value=pointer
      )
    return ProcessedField(field=raw_field)

  def _convert_to_virtual_column(self, field: str) -> VirtualColumn:
    """Converts a field to virtual column."""
    if field.isdigit():
      field = int(field)
    else:
      with contextlib.suppress(ValueError):
        field = float(field)
    if isinstance(field, (int, float)):
      return VirtualColumn(type='built-in', value=field)

    operators = ('/', r'\*', r'\+', ' - ')
    if len(expressions := re.split('|'.join(operators), field)) > 1:
      virtual_column_fields = []
      substitute_expression = field
      for expression in expressions:
        element = expression.strip()
        if self._is_valid_google_ads_field(element):
          virtual_column_fields.append(element)
          substitute_expression = substitute_expression.replace(
            element, f'{{{element}}}'
          )
      return VirtualColumn(
        type='expression',
        value=field.format(**self.macros) if self.macros else field,
        fields=virtual_column_fields,
        substitute_expression=substitute_expression.replace('.', '_'),
      )
    if not self._is_quoted_string(field):
      raise exceptions.GaarfFieldException(
        f"Incorrect field '{field}' in the query '{self.text}'."
      )
    field = field.replace("'", '').replace('"', '')
    field = field.format(**self.macros) if self.macros else field
    return VirtualColumn(type='built-in', value=field)

  def _is_valid_google_ads_field(self, field: str) -> bool:
    """Checks whether field is is a valid Google Ads field."""
    try:
      operator.attrgetter(field)(self.base_client.google_ads_row)
      return True
    except AttributeError:
      return False

  def _extract_resource_element(self, line_elements: str) -> list[str]:
    return re.split('~', line_elements)

  def _extract_pointer(self, line_elements: str) -> list[str]:
    return re.split('->', line_elements)

  def _extract_nested_resource(self, line_elements: str) -> list[str]:
    return re.split(':', line_elements)

  def _format_type_field_name(self, field_name: str) -> str:
    return re.sub(r'\.type', '.type_', field_name)

  def _normalize_column_name(self, column_name: str) -> str:
    return re.sub(r'\.', '_', column_name)

  def _remove_trailing_comma(self, query: str) -> str:
    return re.sub(r',\s+from', ' FROM', query, re.IGNORECASE)

  def _unformat_type_field_name(self, query: str) -> str:
    return re.sub(r'\.type_', '.type', query)

  def _is_quoted_string(self, field_name: str) -> bool:
    if (field_name.startswith("'") and field_name.endswith("'")) or (
      field_name.startswith('"') and field_name.endswith('"')
    ):
      return True
    return False



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import datetime

import pytest

from gaarf import exceptions, query_editor


@pytest.fixture
def query():
  query = """
-- Comment
# Comment
// Comment

SELECT
    1 AS constant,
    '2023-01-01' AS date,
    '{current_date}' AS current_date,
    metrics.clicks / metrics.impressions AS ctr,
    customer.id, --customer_id
    campaign.bidding_strategy_type AS campaign_type, campaign.id:nested AS campaign,
    ad_group.id~1 AS ad_group,
    ad_group_ad.ad.id->asset AS ad,
    metrics.cost_micros * 1e6 AS cost,
    {% if selective == "true" %}
        campaign.selective_optimization AS selective_optimization,
    {% endif %}
from ad_group_ad;
"""
  return query


@pytest.fixture
def builtin_query():
  return 'SELECT * FROM builtin.ocid_mapping'


def test_builtin_query_returns_valid_specification(builtin_query):
  spec = query_editor.QuerySpecification(
    title='/tmp/ocid_mapping.sql', text=builtin_query, args=None
  ).generate()
  assert spec.is_builtin_query
  assert spec.query_title == 'ocid_mapping'
  assert spec.resource_name == 'builtin.ocid_mapping'
  assert not spec.column_names


@pytest.fixture
def query_specification(query):
  return query_editor.QuerySpecification(
    title='sample_query', text=query, args=None
  )


@pytest.fixture
def query_specification_template(query):
  return query_editor.QuerySpecification(
    title='templated_query',
    text=query,
    args={'template': {'selective': 'true'}},
  )


@pytest.fixture
def sample_query(query_specification):
  return query_specification.generate()


@pytest.fixture
def templated_query(query_specification_template):
  return query_specification_template.generate()


class TestRegularQuery:
  def test_correct_title(self, sample_query):
    assert sample_query.query_title == 'sample_query'

  def test_extract_resource_form_query_returns_found_resource(
    self, query_specification
  ):
    resource_name = query_specification._extract_resource_from_query()
    expected_resource_name = 'ad_group_ad'

    assert resource_name == expected_resource_name

  def test_extract_correct_fields(self, sample_query):
    assert sample_query.fields == [
      'customer.id',
      'campaign.bidding_strategy_type',
      'campaign.id',
      'ad_group.id',
      'ad_group_ad.ad.id',
    ]

  def test_extract_correct_aliases(self, sample_query):
    assert sample_query.column_names == [
      'constant',
      'date',
      'current_date',
      'ctr',
      'customer_id',
      'campaign_type',
      'campaign',
      'ad_group',
      'ad',
      'cost',
    ]

  def test_extract_correct_text(self, sample_query):
    assert sample_query.query_text.lower() == (
      'select customer.id, campaign.bidding_strategy_type, campaign.id, '
      'ad_group.id, ad_group_ad.ad.id, metrics.clicks, '
      'metrics.impressions, metrics.cost_micros from ad_group_ad'
    )

  def test_extract_filters_returns_correct_match(self):
    query = query_editor.QuerySpecification(
      text='SELECT campaign.id FROM campaign WHERE campaign.status = ENABLED'
    )
    where_statement = query._extract_filters()
    assert 'WHERE campaign.status = ENABLED' in where_statement

  def test_extract_filters_returns_limit(self):
    query = query_editor.QuerySpecification(
      text='SELECT campaign.id FROM campaign LIMIT 10'
    )
    where_statement = query._extract_filters()
    assert 'LIMIT 10' in where_statement

  def test_extract_filters_returns_order_by(self):
    query = query_editor.QuerySpecification(
      text='SELECT campaign.id FROM campaign ORDER BY 1'
    )
    where_statement = query._extract_filters()
    assert 'ORDER BY 1' in where_statement

  def test_extract_filters_returns_nothing(self):
    query = query_editor.QuerySpecification(
      text='SELECT campaign.id FROM campaign'
    )
    where_statement = query._extract_filters()
    assert where_statement == ''

  def test_extract_custom_callers(self, sample_query):
    assert sample_query.customizers == {
      'campaign': {'type': 'nested_field', 'value': 'nested'},
      'ad_group': {'type': 'resource_index', 'value': 1},
      'ad': {'type': 'pointer', 'value': 'asset'},
    }

  def test_extract_correct_resource(self, sample_query):
    assert sample_query.resource_name == 'ad_group_ad'

  def test_is_constant_resource(self, sample_query):
    assert not sample_query.is_constant_resource

  def test_has_virtual_columns(self, sample_query):
    assert sample_query.virtual_columns == {
      'constant': query_editor.VirtualColumn(type='built-in', value=1),
      'date': query_editor.VirtualColumn(type='built-in', value='2023-01-01'),
      'current_date': query_editor.VirtualColumn(
        type='built-in', value=datetime.date.today().strftime('%Y-%m-%d')
      ),
      'ctr': query_editor.VirtualColumn(
        type='expression',
        value='metrics.clicks / metrics.impressions',
        fields=['metrics.clicks', 'metrics.impressions'],
        substitute_expression='{metrics_clicks} / {metrics_impressions}',
      ),
      'cost': query_editor.VirtualColumn(
        type='expression',
        value='metrics.cost_micros * 1e6',
        fields=['metrics.cost_micros'],
        substitute_expression='{metrics_cost_micros} * 1e6',
      ),
    }

  def test_incorrect_resource_raises_value_error(self):
    query = 'SELECT metrics.clicks FROM ad_groups'
    spec = query_editor.QuerySpecification(
      title='sample_query', text=query, args=None
    )
    with pytest.raises(exceptions.GaarfResourceException):
      spec.generate()

  def test_generate_works_with_virtual_column(self):
    query = (
      'SELECT "something:anything" AS virtual, metrics.clicks FROM ad_group'
    )
    spec = query_editor.QuerySpecification(
      title='sample_query', text=query, args=None
    )
    spec.generate()

  def test_incorrect_specification_raises_macro_error(self):
    query = "SELECT '${custom_field}', ad_group.id FROM ad_group"
    spec = query_editor.QuerySpecification(
      title='sample_query', text=query, args=None
    )
    with pytest.raises(exceptions.GaarfMacroException):
      spec.generate()

  def test_incorrect_specification_raises_virtual_column_error(self):
    query = 'SELECT 1, ad_group.id AS ad_group_id FROM ad_group'
    spec = query_editor.QuerySpecification(
      title='sample_query', text=query, args=None
    )
    with pytest.raises(exceptions.GaarfVirtualColumnException):
      spec.generate()

  def test_incorrect_field_raises_value_error(self):
    query = 'SELECT metric.impressions, ad_group.id FROM ad_group'
    spec = query_editor.QuerySpecification(
      title='sample_query', text=query, args=None
    )
    with pytest.raises(exceptions.GaarfFieldException):
      spec.generate()


class TestTemplatedQuery:
  def test_extract_correct_fields(self, templated_query):
    assert templated_query.fields == [
      'customer.id',
      'campaign.bidding_strategy_type',
      'campaign.id',
      'ad_group.id',
      'ad_group_ad.ad.id',
      'campaign.selective_optimization',
    ]

  def test_extract_correct_aliases(self, templated_query):
    assert templated_query.column_names == [
      'constant',
      'date',
      'current_date',
      'ctr',
      'customer_id',
      'campaign_type',
      'campaign',
      'ad_group',
      'ad',
      'cost',
      'selective_optimization',
    ]

  def test_extract_correct_text(self, templated_query):
    assert templated_query.query_text.lower() == (
      'select customer.id, campaign.bidding_strategy_type, campaign.id, '
      'ad_group.id, ad_group_ad.ad.id, campaign.selective_optimization, '
      'metrics.clicks, metrics.impressions, metrics.cost_micros '
      'from ad_group_ad'
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data with console."""

from __future__ import annotations

import rich
from rich import console, table
from rich import json as rich_json

from gaarf import report as gaarf_report
from gaarf.io.writers import abs_writer


class ConsoleWriter(abs_writer.AbsWriter):
  """Writes reports to standard output.

  Attributes:
    page_size: How many row of report should be written
    type: Type of output ('table', 'json').
  """

  def __init__(
    self, page_size: int = 10, format: str = 'table', **kwargs: str
  ) -> None:
    """Initializes ConsoleWriter.

    Args:
        page_size: How many row of report should be written
        format: Type of output ('table', 'json').
        kwargs: Optional parameter to initialize writer.
    """
    super().__init__(**kwargs)
    self.page_size = int(page_size)
    self.format = format

  def write(self, report: gaarf_report.GaarfReport, destination: str) -> None:
    """Writes Gaarf report to standard output.

    Args:
      report: Gaarf report.
      destination: Base file name report should be written to.
    """
    report = self.format_for_write(report)
    if self.format == 'table':
      output_table = table.Table(
        title=f"showing results for query <{destination.split('/')[-1]}>",
        caption=(
          f'showing rows 1-{min(self.page_size, len(report))} '
          f'out of total {len(report)}'
        ),
        box=rich.box.MARKDOWN,
      )
      for header in report.column_names:
        output_table.add_column(header)
      for i, row in enumerate(report):
        if i < self.page_size:
          output_table.add_row(*[str(field) for field in row])
    elif self.format == 'json':
      output_table = rich_json.JSON(report.to_json())
    console.Console().print(output_table)



from __future__ import annotations

import pandas as pd
import pytest
import sqlalchemy

from gaarf.executors import sql_executor


class TestSqlAlchemyQueryExecutor:
  @pytest.fixture
  def engine(self):
    return sqlalchemy.create_engine('sqlite:///:memory:')

  @pytest.fixture
  def executor(self, engine):
    return sql_executor.SqlAlchemyQueryExecutor(engine)

  def test_execute_returns_data_saved_to_db(self, executor, engine):
    query = 'CREATE TABLE test AS SELECT 1 AS one;'
    executor.execute(script_name='test', query_text=query)

    with engine.connect() as connection:
      result = connection.execute(sqlalchemy.text('select one from test'))
      for row in result:
        assert row.one == 1

  def test_execute_returns_data_to_caller(self, executor):
    query = 'SELECT 1 AS one;'
    expected_result = pd.DataFrame(data=[[1]], columns=['one'])
    result = executor.execute(script_name='test', query_text=query)
    assert result.equals(expected_result)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data to a file."""

import os

from gaarf.io.writers.abs_writer import AbsWriter


class FileWriter(AbsWriter):
  """Writes Gaarf Report to a local or remote file.

  Attributes:
      destination_folder: Destination where output file is stored.
  """

  def __init__(
    self,
    destination_folder: str | os.PathLike = os.getcwd(),
    **kwargs: str,
  ) -> None:
    """Initializes FileWriter based on destination folder."""
    super().__init__(**kwargs)
    self.destination_folder = str(destination_folder)

  def create_dir(self) -> None:
    """Creates folders if needed or destination is not remote."""
    if (
      not os.path.isdir(self.destination_folder)
      and '://' not in self.destination_folder
    ):
      os.makedirs(self.destination_folder)

  def write(self) -> None:
    return



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for generating simulated GaarfReport based on query text."""

from __future__ import annotations

import dataclasses
import random
import string
from collections.abc import Sequence
from typing import Any

from faker import Faker

from gaarf.api_clients import GOOGLE_ADS_API_VERSION, BaseClient
from gaarf.query_editor import QuerySpecification
from gaarf.query_executor import AdsReportFetcher
from gaarf.report import GaarfReport


@dataclasses.dataclass(frozen=True)
class FakeField:
  type: Any
  name: str


@dataclasses.dataclass
class SimulatorSpecification:
  """Specifies meta information for simulated results."""

  api_version: str = GOOGLE_ADS_API_VERSION
  n_rows: int = 1000
  days_ago: str = '-7d'
  string_length: int = 3
  ignored_enums: tuple[str, ...] = ('REMOVED', 'UNKNOWN', 'UNSPECIFIED')
  allowed_enums: dict[str, Sequence[str]] | None = None
  replacements: dict[str, Any] | None = None


def simulate_data(
  query_text: str,
  query_name: str,
  args: dict,
  api_version: str,
  simulator_specification: SimulatorSpecification,
) -> GaarfReport | None:
  """Simulates GaarfReport based for a given query.

  Args:
      query_text: GAQL query text.
      query_name: Name of the query.
      args: Query parameters to be replaced in query_text.
      api_version: Ads API version to simulate against.
      simulator_specification: Meta information for simulated results.

  Returns:
      Report with simulated data. For `built-in` queries returns None.
  """
  query_specification = QuerySpecification(
    query_text, query_name, args, api_version
  ).generate()
  if query_specification.is_builtin_query:
    return None
  client = BaseClient(api_version)
  report_fetcher = AdsReportFetcher(client)
  inferred_types = client.infer_types(query_specification.fields)
  report = report_fetcher.fetch(query_specification)
  try:
    iter(report.results_placeholder[0])
    results = report.results_placeholder[0]
  except TypeError:
    results = [report.results_placeholder[0]]
  if not results:
    results = [results]
  row_field_types = [
    FakeField(type(field_value), field_name)
    for field_value, field_name in zip(results, query_specification.fields)
  ]
  values = _simulate_values(
    simulator_specification,
    query_specification,
    row_field_types,
    inferred_types,
  )
  return GaarfReport(values, query_specification.column_names)


def _simulate_values(
  simulator_specification: SimulatorSpecification,
  query_specification: QuerySpecification,
  row_field_types: list,
  inferred_types: list,
) -> list[list]:
  """Simulates values for a given query based on API version.

  Args:
      simulator_specification: Meta information for simulated results.
      query_specification: Specification of the query.
      row_field_types: Mapping between type of each row element and its name.
      inferred_types: Allowed types for each field of the query.

  Returns:
      Nested list with simulated values.
  """
  simulation_helper = Faker()
  values = []
  for _ in range(simulator_specification.n_rows):
    row = []
    for i, field in enumerate(row_field_types):
      if len(enums := inferred_types[i].values) > 1:
        if allowed_enums := simulator_specification.allowed_enums:
          if selected_enums := allowed_enums.get(field.name):
            value = random.choice(selected_enums)
          else:
            value = random.choice(list(enums))
        else:
          value = random.choice(
            [
              value
              for value in inferred_types[i].values
              if value not in simulator_specification.ignored_enums
            ]
          )
      elif (
        simulator_specification.replacements
        and field.name in simulator_specification.replacements
      ):
        value = random.choice(simulator_specification.replacements[field.name])
      else:
        value = _generate_random_value(
          field,
          simulator_specification,
          simulation_helper,
          query_specification.column_names[i],
        )
      row.append(value)
    values.append(row)
  return values


def _generate_random_value(
  field: FakeField,
  simulator_specification: SimulatorSpecification,
  simulation_helper: Faker,
  column_name: str,
) -> bool | str | float:
  """Generates random values based on field type and name.

  Args:
      field: A single element of a row.
      simulator_specification: Meta information for simulated results.
      simulation_helper: Class to perform simulation operations.
      column_name: Name of the column that needs to be simulated.

  Returns:
      Simulated value.
  """
  if field.type == str:
    if 'date' in field.name:
      return simulation_helper.date_between(
        simulator_specification.days_ago
      ).strftime('%Y-%m-%d')
    if 'url' in field.name:
      return 'example.com'
    if 'video.id' in field.name or 'video_id' in field.name:
      return '4WXs3sKu41I'
    if 'id' in column_name:
      return str(random.randint(1000000, 1000010))
    return _generate_random_string(simulator_specification.string_length)
  if field.type == int:
    if '.id' in field.name:
      return random.randint(1000000, 1000010)
    if 'micros' in field.name:
      return int(random.randint(0, 1000) * 1e6)
    return random.randint(0, 1000)
  if field.type == float:
    return float(random.randint(0, 1000))
  if field.type == bool:
    return bool(random.getrandbits(1))
  return ''


def _generate_random_string(length: int) -> str:
  'Helper for simulating random string of a given length.' ''
  characters = string.ascii_lowercase
  return ''.join(random.choice(characters) for _ in range(length))
